###############################################################################
#                                                                             #
# Bricklayer - An Installer for IPFire                                        #
# Copyright (C) 2021 IPFire Development Team                                  #
#                                                                             #
# This program is free software; you can redistribute it and/or               #
# modify it under the terms of the GNU General Public License                 #
# as published by the Free Software Foundation; either version 2              #
# of the License, or (at your option) any later version.                      #
#                                                                             #
# This program is distributed in the hope that it will be useful,             #
# but WITHOUT ANY WARRANTY; without even the implied warranty of              #
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the               #
# GNU General Public License for more details.                                #
#                                                                             #
# You should have received a copy of the GNU General Public License           #
# along with this program.  If not, see <http://www.gnu.org/licenses/>.       #
#                                                                             #
###############################################################################

import itertools
import logging

from . import step
from .i18n import _, N_

# Setup logging
log = logging.getLogger("bricklayer.bootloaders")

class Bootloader(object):
	"""
		A generic class for bootloaders
	"""

	# Set the name of the bootloader
	name = None

	# Which architectures are supported?
	arches = []

	# Packages that need to be installed
	packages = []

	# Requires a BIOS boot partition?
	requires_bootldr_partition = False

	# Requires an EFI partition?
	requires_efi_partition = False

	def __init__(self, bricklayer):
		self.bricklayer = bricklayer

	@property
	def supported(self):
		"""
			Returns True if this booloader is supported
		"""
		# If arches is empty, this bootloader is supported
		if not self.arches:
			return True

		return self.bricklayer.arch in self.arches

	def install(self):
		"""
			Called to install the bootloader
		"""
		pass


class Grub(Bootloader):
	name = N_("GRUB")

	arches = [
		"x86_64",
	]

	packages = [
		"grub",
	]

	requires_bootldr_partition = True

	def install(self):
		# Initialize Pakfire
		pakfire = self.bricklayer.setup_pakfire()

		# Install GRUB2 on all disks that have been selected
		for disk in self.bricklayer.disks.selected:
			self.install_on_disk(disk, pakfire)

	@property
	def grub_arch(self):
		return "i386-pc"

	@property
	def grub_args(self):
		return [
			"--target=%s" % self.grub_arch,
		]

	def install_on_disk(self, disk, pakfire):
		log.info("Installing GRUB on %s" % disk.path)

		pakfire.execute(["grub-install"] + self.grub_args + [disk.path])


class GrubEFI(Grub):
	name = N_("GRUB EFI")

	requires_efi_partition = True

	@property
	def grub_arch(self):
		if self.bricklayer.arch == "aarch64":
			return "arm64-efi"

		elif self.bricklayer.arch == "x86_64":
			return "x86_64-efi"

	@property
	def grub_args(self):
		return [
			"--target=%s" % self.grub_arch,
			"--efi-directory=/boot/efi",
		]


class InstallBootloader(step.Step):
	"""
		Installs the bootloader
	"""
	def run(self):
		# Find all packages that need to be installed
		packages = []
		for bootloader in self.bricklayer.bootloaders:
			packages += bootloader.packages

		# Install them
		self.install_packages(packages)

		# Install all bootloaders
		for bootloader in self.bricklayer.bootloaders:
			with self.tui.progress(
				_("Installing Bootloader"),
				_("Installing bootloader \"%s\"...") % _(bootloader.name),
			):
				bootloader.install()
