###############################################################################
#                                                                             #
# Bricklayer - An Installer for IPFire                                        #
# Copyright (C) 2021 IPFire Development Team                                  #
#                                                                             #
# This program is free software; you can redistribute it and/or               #
# modify it under the terms of the GNU General Public License                 #
# as published by the Free Software Foundation; either version 2              #
# of the License, or (at your option) any later version.                      #
#                                                                             #
# This program is distributed in the hope that it will be useful,             #
# but WITHOUT ANY WARRANTY; without even the implied warranty of              #
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the               #
# GNU General Public License for more details.                                #
#                                                                             #
# You should have received a copy of the GNU General Public License           #
# along with this program.  If not, see <http://www.gnu.org/licenses/>.       #
#                                                                             #
###############################################################################

import logging
import os
import time
import pakfire.errors

import snack

from . import i18n
from .i18n import _

# Setup logging
log = logging.getLogger("bricklayer.step")

class Step(object):
	"""
		Steps are the heart of this installer. They perform actions and are
		being executed one after the other.
	"""
	# This enables or disables this step
	enabled = True

	# Should this be run in first-install mode?
	first_install = False

	def __init__(self, bricklayer, tui):
		self.bricklayer = bricklayer
		self.tui = tui

		log.debug("Initializing step %s" % self.__class__.__name__)

		# Call the custom initialization
		self.initialize()

	def initialize(self):
		"""
			Custom initialization action to be overlayed
		"""
		pass

	def run(self):
		"""
			Run this step - to be overlayed
		"""
		pass

	def install_packages(self, packages):
		# Nothing to do if there are no packages
		if not packages:
			return

		# Set up Pakfire
		with self.tui.progress(
			_("Setting Up Pakfire"),
			_("Pakfire is being set up..."),
		):
			p = self.bricklayer.setup_pakfire()

		with p as p:
			# Resolve package dependencies
			with self.tui.progress(
				_("Resolving Dependencies"),
				_("Resolving package dependencies..."),
			):
				try:
					transaction = p.install(packages)

				# Abort on any dependencies problems
				except pakfire.errors.DependencyError as e:
					problems = e.args[0]

					# Format problem descriptions
					text = []
					for problem in problems:
						lines = [
							"* %s" % problem
						]
						for solution in problem.solutions:
							lines.append("  --> %s" % solution)

						text.append("\n".join(lines))

					self.tui.error(
						_("Dependency Problem"),
						_(
							"A problem has occured during resolving package dependencies:\n\n%s",
							"Problems have occured during resolving package dependencies:\n\n%s",
							len(problems),
						) % "\n\n".join(text),
						width=78,
					)

			# Log the transaction
			log.info("%s" % transaction.dump())

			# Run the transaction
			with self.tui.progress(
				_("Installing Packages"),
				_("Installing packages..."),
			):
				transaction.run()


class InteractiveStep(Step):
	"""
		A convenience handler that is disabled in unattended mode
	"""
	@property
	def enabled(self):
		# Disable in unattended mode
		return not self.bricklayer.unattended


class Welcome(InteractiveStep):
	"""
		Shows a very warm welcome message to the user
	"""
	first_install = True

	def run(self):
		name = self.bricklayer.os.get("NAME")
		current_language = self.bricklayer.settings.get("language")

		# Let the user select
		lang = self.tui.select(
			_("Willkommen, Bienvenue, Welcome!"),
			_("Select the language you wish to use for the installation"),
			i18n.supported_languages, default=current_language,
			buttons=[_("Start Installation")], width=60,
		)

		log.info("Language selected: %s" % lang)

		# Store in settings
		self.bricklayer.settings["language"] = lang

		# Set to environment
		os.environ["LANGUAGE"] = lang


class Congratulations(InteractiveStep):
	"""
		Shows a message that the installation is complete
	"""
	first_install = True

	def run(self):
		self.tui.message(
			_("Congratulations"),
			_(
				"The installation has been completed successfully."
				"\n\n"
				"You can now safely remove any media used for the installation."
			),
			buttons=[_("Finish")],

			# Make the window a litte bit bigger
			width=50,
		)


class UnattendedWarning(Step):
	@property
	def enabled(self):
		# Only enabled in unattended mode
		return self.bricklayer.unattended

	def run(self):
		seconds = 10

		p = self.tui.progress(
			_("Unattended Installation"),
			_("Unattended installation is starting in %s seconds") % seconds,
			max_value=seconds * 10,
		)

		with p:
			for i in range(seconds * 10):
				time.sleep(0.1)

				p.update(i + 1)


class RootPassword(InteractiveStep):
	first_install = True

	def run(self):
		password = self.tui.passwd(
			_("Root Password"),
			_("Please enter the password for the 'root' user"),
		)

		# Save the password for now
		self.bricklayer.settings["root-password"] = password


class ToggleFirstInstallStatus(Step):
	first_install = True

	@property
	def enabled(self):
		# Only enabled in unattended and first install mode
		return self.bricklayer.unattended or self.bricklayer.first_install

	def run(self):
		# Path to the marker file
		path = os.path.join(self.bricklayer.root, ".firstinstall")

		# Create file in unattended mode
		if self.bricklayer.unattended:
			with open(path, "w"):
				pass

		# Remove the file in first install mode
		elif self.bricklayer.first_install:
			try:
				os.unlink(path)
			except FileNotFoundError:
				pass
